import psutil
from rich import box
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from textual.widget import Widget

from ._helpers import sizeof_fmt


def get_process_list(num_procs: int):
    processes = list(
        psutil.process_iter(
            [
                "pid",
                "name",
                "username",
                "cmdline",
                "cpu_percent",
                "num_threads",
                "memory_info",
                "status",
            ]
        )
    )

    if processes[0].pid == 0:
        # Remove process with PID 0. On Windows, that's SYSTEM IDLE, and we
        # don't want that to appear at the top of the list.
        # <https://twitter.com/andre_roberge/status/1488885893716975622/photo/1>
        processes = processes[1:]

    processes = [p.info for p in processes]

    processes = sorted(
        processes,
        # The item.info["cpu_percent"] can be `ad_value` (default None).
        # It gets assigned to a dict key in case AccessDenied or
        # ZombieProcess exception is raised when retrieving that particular
        # process information.
        key=lambda p: (p["cpu_percent"] or 0.0),
        reverse=True,
    )
    processes = processes[:num_procs]
    return processes


class ProcsList(Widget):
    def on_mount(self):
        self.max_num_procs = 100
        self.collect_data()
        self.set_interval(6.0, self.collect_data)

    def collect_data(self):
        processes = get_process_list(self.max_num_procs)

        table = Table(
            show_header=True,
            header_style="bold",
            box=None,
            padding=(0, 1),
            expand=True,
        )
        # set ration=1 on all columns that should be expanded
        # <https://github.com/Textualize/rich/issues/2030>
        table.add_column(Text("pid", justify="left"), no_wrap=True, justify="right")
        table.add_column("program", style="green", no_wrap=True, ratio=1)
        table.add_column("args", no_wrap=True, ratio=2)
        table.add_column(
            Text("thr", justify="left"),
            style="green",
            no_wrap=True,
            justify="right",
        )
        table.add_column("user", no_wrap=True)
        table.add_column(
            Text("mem", justify="left"), style="green", no_wrap=True, justify="right"
        )
        table.add_column(
            Text("cpu%", style="u", justify="left"),
            no_wrap=True,
            justify="right",
        )

        for p in processes:
            # Everything can be None here, see comment above
            pid = p["pid"]
            pid = "" if pid is None else str(pid)
            #
            name = p["name"]
            if name is None:
                name = ""
            #
            cmdline = p["cmdline"]
            cmdline = "" if cmdline is None else " ".join(p["cmdline"][1:])
            #
            num_threads = p["num_threads"]
            num_threads = "" if num_threads is None else str(num_threads)
            #
            username = p["username"]
            if username is None:
                username = ""
            #
            mem_info = p["memory_info"]
            mem_info = (
                "" if mem_info is None else sizeof_fmt(mem_info.rss, suffix="", sep="")
            )
            #
            cpu_percent = p["cpu_percent"]
            cpu_percent = "" if cpu_percent is None else f"{cpu_percent:.1f}"
            table.add_row(
                pid, name, cmdline, num_threads, username, mem_info, cpu_percent
            )

        total_num_threads = sum((p["num_threads"] or 0) for p in processes)
        num_sleep = sum(p["status"] == "sleeping" for p in processes)

        self.panel = Panel(
            table,
            title=f"[b]proc[/] - {len(processes)} ({total_num_threads} thr), {num_sleep} slp",
            title_align="left",
            # border_style="cyan",
            border_style="white",
            box=box.SQUARE,
        )

        self.refresh()

    def render(self) -> Panel:
        return self.panel

    async def on_resize(self, event):
        self.max_num_procs = event.height - 3



import psutil
from rich import box
from rich.console import Group
from rich.panel import Panel
from rich.text import Text
from textual.widget import Widget

from ._helpers import sizeof_fmt
from .braille_stream import BrailleStream


class Mem(Widget):
    def on_mount(self):
        mem = psutil.virtual_memory()
        self.mem_total_bytes = mem.total

        # check which mem sections are available on the machine
        self.attrs = []
        self.colors = ["yellow", "green", "blue", "magenta", "red"]
        for attr in ["free", "available", "cached", "used"]:
            if hasattr(mem, attr):
                self.attrs.append(attr)

        swap = psutil.swap_memory()
        if swap is not None:
            self.attrs.append("swap")

        # append spaces to make all names equally long
        maxlen = max(len(string) for string in self.attrs)
        maxlen = min(maxlen, 5)
        self.labels = [attr[:maxlen].ljust(maxlen) for attr in self.attrs]

        # can't use
        # [BrailleStream(40, 4, 0.0, self.mem_total_bytes)] * len(self.names)
        # since that only creates one BrailleStream, references n times.
        self.mem_streams = []
        for attr in self.attrs:
            total = swap.total if attr == "swap" else self.mem_total_bytes
            self.mem_streams.append(BrailleStream(40, 4, 0.0, total))

        self.group = Group("", "", "", "", "")

        mem_total_string = sizeof_fmt(self.mem_total_bytes, fmt=".2f")
        self.panel = Panel(
            self.group,
            title=f"[b]mem[/] - {mem_total_string}",
            title_align="left",
            border_style="white",
            box=box.SQUARE,
        )

        self.refresh_table()
        self.set_interval(2.0, self.refresh_table)

    def refresh_table(self):
        mem = psutil.virtual_memory()
        swap = psutil.swap_memory()

        for k, (attr, label, stream, col) in enumerate(
            zip(self.attrs, self.labels, self.mem_streams, self.colors)
        ):
            if attr == "swap":
                val = swap.used
                total = swap.total
                if total == 0:
                    total = 1
            else:
                val = getattr(mem, attr)
                total = self.mem_total_bytes

            stream.add_value(val)
            val_string = " ".join(
                [
                    label,
                    sizeof_fmt(val, fmt=".2f"),
                    f"({val / total * 100:.0f}%)",
                ]
            )
            graph = "\n".join(
                [val_string + stream.graph[0][len(val_string) :]] + stream.graph[1:]
            )
            self.group.renderables[k] = Text(graph, style=col)

        self.refresh()

    def render(self) -> Panel:
        return self.panel

    async def on_resize(self, event):
        for ms in self.mem_streams:
            ms.reset_width(event.width - 4)

        # split the available event.height-2 into n even blocks, and if there's
        # a rest, divide it up into the first, e.g., with n=4
        # 17 -> 5, 4, 4, 4
        n = len(self.attrs)
        heights = [(event.height - 2) // n] * n
        for k in range((event.height - 2) % n):
            heights[k] += 1
            # add to last:
            # heights[-(k + 1)] += 1

        for ms, h in zip(self.mem_streams, heights):
            ms.reset_height(h)

        self.refresh_table()



import psutil
from rich import box
from rich.panel import Panel
from rich.text import Text
from textual.widget import Widget

from .braille_stream import BrailleStream


class Battery(Widget):
    def on_mount(self):
        self.bat_stream = BrailleStream(40, 3, 0.0, 100.0)

        self.panel = Panel(
            "",
            title="",
            title_align="left",
            # border_style="yellow",
            border_style="white",
            box=box.SQUARE,
        )
        self.collect_data()
        # update frequency: 1 min
        self.set_interval(60.0, self.collect_data)

    def collect_data(self):
        bat = psutil.sensors_battery()
        assert bat is not None

        self.bat_stream.add_value(bat.percent)
        self.refresh_graph()

        if bat.power_plugged:
            status = "charging"
        else:
            mm = bat.secsleft // 60
            hh, mm = divmod(mm, 60)
            time_left_str = []
            if hh > 0:
                time_left_str.append(f"{hh}h")
            if mm > 0:
                time_left_str.append(f"{mm}min")
            status = " ".join(time_left_str) + " left"

        title = f"[b]battery[/] - {self.bat_stream.values[-1]:.1f}% - {status}"
        if bat.percent < 15 and not bat.power_plugged:
            title = "[red reverse bold]" + title + "[/]"

        self.panel.title = title

        self.refresh()

    def refresh_graph(self):
        self.panel.renderable = Text("\n".join(self.bat_stream.graph), style="yellow")

    def render(self) -> Panel:
        return self.panel

    async def on_resize(self, event):
        self.bat_stream.reset_width(event.width - 4)
        self.bat_stream.reset_height(event.height - 2)
        self.refresh_graph()



from __future__ import annotations

import socket

import psutil
from rich import box
from rich.console import Group
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from textual.widget import Widget

from .__about__ import __version__
from ._helpers import sizeof_fmt
from .braille_stream import BrailleStream

# def get_ip():
#     s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
#     s.connect(("8.8.8.8", 80))
#     ip = s.getsockname()[0]
#     s.close()
#     return ip


def _autoselect_interface():
    """try to find non-lo and non-docker interface that is up"""
    stats = psutil.net_if_stats()
    score_dict = {}
    for name, stats in stats.items():
        if not stats.isup:
            score_dict[name] = 0
        elif (
            # On Unix, we have `lo`, on Windows `Loopback Pseudo-Interface k`
            # and `Local Area Connection k` (the latter is valid)
            name.startswith("lo")
            or name.lower().startswith("loopback")
            or name.lower().startswith("docker")
        ):
            score_dict[name] = 1
        elif name.lower().startswith("fw") or name.lower().startswith("Bluetooth"):
            # firewire <https://github.com/nschloe/tiptop/issues/45#issuecomment-991884364>
            # or bluetooth
            score_dict[name] = 2
        else:
            score_dict[name] = 3

    # Amongst all keys with max score, get the alphabetically first.
    # This is to prefer en0 over en5, <https://github.com/nschloe/tiptop/issues/81>.
    max_score = max(score_dict.values())
    max_keys = [key for key, score in score_dict.items() if score == max_score]
    return sorted(max_keys)[0]


class Net(Widget):
    def __init__(self, interface: str | None = None):
        self.interface = _autoselect_interface() if interface is None else interface
        self.tiptop_string = f"tiptop v{__version__}"
        super().__init__()

    def on_mount(self):
        self.down_box = Panel(
            "",
            title="▼ down",
            title_align="left",
            style="green",
            width=20,
            box=box.SQUARE,
        )
        self.up_box = Panel(
            "",
            title="▲ up",
            title_align="left",
            style="blue",
            width=20,
            box=box.SQUARE,
        )
        self.table = Table(expand=True, show_header=False, padding=0, box=None)
        # Add ratio 1 to expand that column as much as possible
        self.table.add_column("graph", no_wrap=True, ratio=1)
        self.table.add_column("box", no_wrap=True, width=20)
        self.table.add_row("", self.down_box)
        self.table.add_row("", self.up_box)

        self.group = Group(self.table, "", "")
        self.panel = Panel(
            self.group,
            title=f"[b]net[/] - {self.interface}",
            # border_style="red",
            border_style="white",
            title_align="left",
            box=box.SQUARE,
            subtitle=self.tiptop_string,
            subtitle_align="right",
        )

        self.last_net = None
        self.max_recv_bytes_s = 0
        self.max_recv_bytes_s_str = ""
        self.max_sent_bytes_s = 0
        self.max_sent_bytes_s_str = ""

        self.recv_stream = BrailleStream(20, 5, 0.0, 1.0e6)
        self.sent_stream = BrailleStream(20, 5, 0.0, 1.0e6, flipud=True)

        self.refresh_ips()
        self.refresh_panel()

        self.interval_s = 2.0
        self.set_interval(self.interval_s, self.refresh_panel)
        self.set_interval(60.0, self.refresh_ips)

    def refresh_ips(self):
        addrs = psutil.net_if_addrs()[self.interface]
        ipv4 = []
        for addr in addrs:
            # ipv4?
            if addr.family == socket.AF_INET:
                ipv4.append(addr.address + " / " + addr.netmask)
        ipv6 = []
        for addr in addrs:
            # ipv4?
            if addr.family == socket.AF_INET6:
                ipv6.append(addr.address)

        ipv4 = "\n      ".join(ipv4)
        ipv6 = "\n      ".join(ipv6)
        self.group.renderables[1] = f"[b]IPv4:[/] {ipv4}"
        self.group.renderables[2] = f"[b]IPv6:[/] {ipv6}"

    # would love to collect data upon each render(), but render is called too often
    # <https://github.com/willmcgugan/textual/issues/162>
    def refresh_panel(self):
        net = psutil.net_io_counters(pernic=True)[self.interface]
        if self.last_net is None:
            recv_bytes_s_string = ""
            sent_bytes_s_string = ""
        else:
            recv_bytes_s = (net.bytes_recv - self.last_net.bytes_recv) / self.interval_s
            recv_bytes_s_string = sizeof_fmt(recv_bytes_s, fmt=".1f") + "/s"
            sent_bytes_s = (net.bytes_sent - self.last_net.bytes_sent) / self.interval_s
            sent_bytes_s_string = sizeof_fmt(sent_bytes_s, fmt=".1f") + "/s"

            if recv_bytes_s > self.max_recv_bytes_s:
                self.max_recv_bytes_s = recv_bytes_s
                self.max_recv_bytes_s_str = sizeof_fmt(recv_bytes_s, fmt=".1f") + "/s"

            if sent_bytes_s > self.max_sent_bytes_s:
                self.max_sent_bytes_s = sent_bytes_s
                self.max_sent_bytes_s_str = sizeof_fmt(sent_bytes_s, fmt=".1f") + "/s"

            self.recv_stream.add_value(recv_bytes_s)
            self.sent_stream.add_value(sent_bytes_s)

        self.last_net = net

        total_recv_string = sizeof_fmt(net.bytes_recv, sep=" ", fmt=".1f")
        total_sent_string = sizeof_fmt(net.bytes_sent, sep=" ", fmt=".1f")

        self.down_box.renderable = "\n".join(
            [
                f"{recv_bytes_s_string}",
                f"max   {self.max_recv_bytes_s_str}",
                f"total {total_recv_string}",
            ]
        )
        self.up_box.renderable = "\n".join(
            [
                f"{sent_bytes_s_string}",
                f"max   {self.max_sent_bytes_s_str}",
                f"total {total_sent_string}",
            ]
        )
        self.refresh_graphs()

        self.refresh()

    def refresh_graphs(self):
        self.table.columns[0]._cells[0] = Text(
            "\n".join(self.recv_stream.graph), style="green"
        )
        self.table.columns[0]._cells[1] = Text(
            "\n".join(self.sent_stream.graph), style="blue"
        )

    def render(self):
        return self.panel

    async def on_resize(self, event):
        self.sent_stream.reset_width(event.width - 25)
        self.recv_stream.reset_width(event.width - 25)
        self.refresh_graphs()
